/* 

 Version 5 - better display layout
 
 Version 4 - now uses TL431 as voltage reference (the TL431 has 1% accuracy of 2.5V reference) actual measured voltages are within 10mv now
    
 Version 3 - changed cutoff voltage to 0.95V (approx) and changed elapsed time to display load current instead. Time still shows when done.
 Now indicates DONE! instead of continuing to display load current when done.
 Got rid of the decimal point in the time display as it was not displaying seconds as 0-60; instead was showing 0-100

 
*/

#include <LiquidCrystal.h>


// Set some constants -- these will need to be adjusted for your setup
//
// Connect a ~10KΩ (Rr) resistor from 3.3V supply pin to the AREF pin,
//   creating a voltage divder with the internal 32KΩ resistor on AREF.
//      aRefVoltage = supply_voltage * 32K / Rr 
//   Alternatively, measure the actual AREF voltage applied with a good DMM:
//  Dev's measured aRefVoltage = 2.62; (mine 2.56 on Uno Nano board)
const float     aRefVoltage = 2.50;

// Connect a load resistor (Rl) to the battery. ~2.2Ω restistor gives ~500mA drain which
//   is about right for a battery rated at 2500mAH.  Note, this should be a >=1W resistor!
//   Dont trust your DMM to meausre such a low resistance accurately. I measured the current
//   and back-calculated the resistance. Best just to trust rated resistance.
//   my  resistance = 2.3;
//
// The integrator works by accumulating the sampled voltage values from the start until
//   hitting a pre-determined low-voltage threshold. Since they are sampled at a 
//   known period, the number of samples taken cancels out and the accumulation of all samples 
//   simply needs to be scalled by a factor
//
//   Let's make some definitions to show the derivation of how this is so:
//   I = current 
//   V = load voltage
//   Rl = load resistance
//   samples = number of voltage samples made
//   sensor = value read out of analog A->D
//   rate = number of samples taken per second
//   
//                                        ave(V)
//   capacity = ∫I ~= ave(I) * time = ----------- * time
//                                          Rl
//
//   Where:
//                ∑(V)     ∑(sensor) * aRefVoltage/1024
//     ave(V) = -------  = ---------------------------------
//              samples              samples
//
//      time  =  samples/rate 
//
//   Thus:
//               ∑(sensor) * aRefVoltage        samples
//   capacity = ----------------------------- * ----------
//                samples * 1024 * Rl            rate
//
//               ∑(sensor) * aRefVoltage 
//   capacity = ------------------------------ = ∑(sensor) * quanta 
//                  1024 * Rl * rate
//
//
//                aRefVoltage
//   quanta =  --------------------  
//               1024 * Rl * rate
//
//   rate in this sketch is 1 Hz (1/sec).  
//   Thus, the units of quanta are: volt*seconds/(Ω*samples) = amp*seconds per sample
//
//   What we really want are mA*Hours, so, scale by 1hr/3600seconds and 1000mA/A, or simply 1/3.6
//
//   So, finally:
//   quanta = aRefVoltage/(Rl*1024*3.6)  in units of mA*hours per sample

const double quanta = 0.00029485; // 2.62/(2.2*1024*3.6) (Mine Aref 2.50, R1 2.3 - Meter probes have 0.3 ohm resistance)


// Define low voltage threshold where any remaining capacity is "unusable"
//        lowThreshold = 0.95V * 1024/aRefVoltage
const int lowThreshold = 389;

int sensorPin          = 0;    // select the analog input pin for the voltage measurement
int ledPin             = 13;   // select the pin for the LED
int fetGatePin         = 8;    // select the pin for the FET
int sensorValue        = 0;    // unscaled sensor output 
float voltage          = 0;    // measured voltage // the AVR has no floating point instruction so floating point math takes 10000+ CPU cycles (slow!)
float loadcurrent      = 0;    // calculated load current // try and find better way of doing this without using floating point math in version 5.
double mAH             = 0;    // calculated mAh
long accumulator       = 0;    // sum of all unscalled sensor values sampled
int epoch              = 0;    // seconds since battery was inserted 
int lowVolts           = 0;    // debounce the low voltage threshold
bool done              = false;// Voltage has dropped below threshold
bool batteryIn         = false;// Battery present

#define R1 2.3; // change this to your load resistors actual value. Also don't forget to re-calculate the quanta as well!

// initialize the the LCD library with the numbers of the interface pins
//  LCD Pins  --  RS, EN, D4, D5, D6, D7
LiquidCrystal lcd(12, 11,  5,  4,  3,  2);


void setup() {
  // declare the ledPin and fetGatePin as an OUTPUT:
  pinMode(ledPin,     OUTPUT);  
  pinMode(fetGatePin, OUTPUT);  

  // set analog reference to external
  analogReference(EXTERNAL);
  Serial.begin(9600);

  // set up the LCD's number of rows and columns: 
  lcd.begin(16, 2);

  lcd.print("2019 A. Smith");
  lcd.setCursor(0, 1);
  lcd.print("S/W ver 1.5");
  delay(2000);
  lcd.setCursor(0, 0);
  lcd.clear();
  lcd.print("Rechargeable"); 
  lcd.setCursor(0, 1);
  lcd.print("Battery Tester");
  delay(2000);
  lcd.clear();  
  lcd.setCursor(0, 0);

  // Print a helpful start-up message to the LCD.
  lcd.print("Insert battery");
  lcd.setCursor(0, 1);
  lcd.print("to start test");
  
  Serial.begin(9600); 
}


void loop() {
  
  sensorValue = analogRead(sensorPin);    


  if (!batteryIn && (sensorValue > 491)) {
    // only start if voltage is over 1.2V. Prevents test from restarting after load disconnection.
    batteryIn   = true;
    done        = false;
    epoch       = 0;
    accumulator = 0;
    lowVolts    = 0;
    digitalWrite(ledPin, LOW);
    digitalWrite(fetGatePin, HIGH);
    // clear out when LCD when starting over
    lcd.clear();
    
  }
  else if (batteryIn && done && (sensorValue <= 20)) {
    // consider the battery removed only after finishing the last measurement to 
    // debounce a glitchy contact
    batteryIn=false; 
  } 
  else if (batteryIn && !done) {
    // Running state during discharge state
 
    voltage = sensorValue*aRefVoltage/1024.0;
    loadcurrent = voltage/R1;
  
    // print voltage
    lcd.setCursor(0, 0);
    lcd.print(voltage);
    lcd.print("V");
    lcd.setCursor(7, 0);
    lcd.print(loadcurrent);
    lcd.print("A RUN");
          
    if (!done) {
      accumulator += sensorValue;
      // print mAH
      mAH = accumulator*quanta;
     
      lcd.setCursor(0,1);
      lcd.print(int(mAH));
      lcd.print("mAh ");
      if  (mAH <   1000) 
      lcd.setCursor(7,1); 
      else
      lcd.setCursor(8,1);
      
      int chours, cminutes, cseconds;
      chours = epoch/60 / 60;
      cminutes = epoch /60 % 60;
      cseconds = epoch %60;
      if (chours <= 9)
      {
      lcd.print("0");
      }
      lcd.print(chours);
      lcd.print(":");
      if (cminutes <= 9)
      {
       lcd.print("0");
      }
       lcd.print(cminutes);
       lcd.print(":");
      if (cseconds <= 9)
      {
       lcd.print("0");
      }
       lcd.print(cseconds);          
    }
        
    if (!done) {
  
      // lowVolts requires 10 seconds in the last 20 before being done
      if (sensorValue < lowThreshold) { lowVolts++;}
      else if (lowVolts > 0)          { lowVolts--;}  
  
      // If it's below threshold for 10 of 20 samples, bail out
      if (lowVolts > 10) {
        done=true;
        digitalWrite(fetGatePin, LOW);
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("Capacity ");
        lcd.print(int(mAH));
        lcd.print(" mAh ");
        lcd.setCursor(0, 1);
        lcd.print("Runtime ");
        
        int chours, cminutes, cseconds;
        chours = epoch/60 / 60;
        cminutes = epoch /60 % 60;
        if (chours <= 9)
        {
         lcd.print("0");
        }
       lcd.print(chours);
       lcd.print("h ");
       if (cminutes <= 9)
        {
         lcd.print("0");
        }
       lcd.print(cminutes);
       lcd.print("m");        
    
      }
    }

    epoch++;
  } // batteryIn -- main routine


  
  if (done) {
    // When done, flash the LED to get attention
    digitalWrite(ledPin, HIGH);
    delay(500);
    digitalWrite(ledPin, LOW);
    delay(499);
  } else {
    // Since the processing takes some time prior to the delay, we'll assume 1mS 
    // This could stand to be improved with an interrupt routine that is kicked off
    // before all the processing starts for each loop
    delay(1000); 
  }

  
}
